from tweetfeels.utils import clean
from tweepy.streaming import StreamListener
from tweepy.utils import parse_datetime
from nltk.sentiment.vader import SentimentIntensityAnalyzer
import json
import time


class Tweet(object):
    """
    Tweet object model. Access to tweet data works like a dict.

    :param data: A dict converted from json string representation of a tweet.
    """
    def __init__(self, data):
        self._data = data
        self._sentiment = None
        self._user_keys = (
            'followers_count', 'friends_count', 'location'
        )
        self._sentiment_keys = (
            'sentiment', 'pos', 'neu', 'neg'
        )
        try:
            ts = parse_datetime(data['created_at'])
            data['created_at'] = ts
        except KeyError:
            print(data)
            raise

    def __len__(self):
        return len(self.keys())

    def __contains__(self, other):
        return other in self.keys()

    def __getitem__(self, key):
        if key in self._user_keys:
            return self._data['user'][key]
        elif key in self._sentiment_keys:
            if key=='sentiment':
                return self.sentiment['compound']
            else:
                return self.sentiment[key]
        else:
            return self._data[key]

    def __str__(self):
        return str({k: self[k] for k in self.keys()})

    @property
    def sentiment(self):
        if self._sentiment is None:
            t = clean(self._data['text'])
            self._sentiment = SentimentIntensityAnalyzer().polarity_scores(t)
        return self._sentiment

    def keys(self):
        k = tuple(self._data.keys())
        k += self._sentiment_keys
        if 'user' in self._data:
            k += self._user_keys
        return k


class TweetListener(StreamListener):
    """
    Expects the controller to implement the handler methods.
    """
    def __init__(self, controller):
        self._controller = controller
        self.waited = 0

    def on_connect(self):
        self.waited = 0

    def on_data(self, data):
        dat = json.loads(data)
        if isinstance(dat, list):
            for d in dat:
                if 'created_at' in d:
                    twt = Tweet(d)
                    if hasattr(self._controller.on_data, '__call__'):
                        self._controller.on_data(twt)
                else:
                    continue
        else:
            if 'created_at' in dat:
                twt = Tweet(dat)
                if hasattr(self._controller.on_data, '__call__'):
                    self._controller.on_data(twt)
        return True

    def on_error(self, status):
        print(status)
        if self.waited == 0:
            if status == 420:
                self.waited = 60
            else:
                self.waited = 5
        self.reconnect_wait('exponential')

        if hasattr(self._controller.on_error, '__call__'):
            ret = self._controller.on_error(status)
        return True

    def reconnect_wait(self, pattern):
        if pattern == 'linear':
            time.sleep(self.waited)
            self.waited += 1
        elif pattern == 'exponential':
            time.sleep(self.waited)
            self.waited *= 2

    def on_disconnect(self, notice):
        """Called when twitter sends a disconnect notice
        Disconnect codes are listed here:
        https://dev.twitter.com/docs/streaming-apis/messages#Disconnect_messages_disconnect
        """
        msg = json.loads(notice)['disconnect']
        if msg['code'] == 4 or msg['code'] > 8:
            self.reconnect_wait('linear')
            self._controller.start()
        else:
            print(f'Disconnected: {msg["code"]}: {msg["reason"]}')



import unittest
from unittest.mock import patch, MagicMock
import json
import os
import time

from tweetfeels import (TweetFeels, Tweet, TweetData)


class Test_Feels(unittest.TestCase):
    def setUp(self):
        TweetFeels._db_factory = (lambda db: MagicMock())
        TweetFeels._auth_factory = (lambda cred: MagicMock())
        TweetFeels._listener_factory = (lambda ctrl: MagicMock())
        TweetFeels._stream_factory = (lambda auth, listener: MagicMock())
        self.tweets_data_path = 'test/sample.json'

    def test_start(self):
        mock_feels = TweetFeels("abcd")
        mock_feels.tracking = []
        mock_feels.start()
        mock_feels._stream.filter.assert_not_called()
        mock_feels.tracking = ['tsla']
        mock_feels.start()
        mock_feels._stream.filter.assert_called_once()

    def test_stop(self):
        mock_feels = TweetFeels("abcd")
        mock_feels.stop()
        mock_feels._stream.disconnect.assert_called_once()

    def test_on_data(self):
        mock_feels = TweetFeels("abcd")
        mock_feels.buffer_limit = 0
        data = {'filter_level': 'low', 'text': 'test data'}
        mock_feels.on_data(data)
        mock_feels._feels.insert_tweet.assert_called_once()

        # test filtering levels
        mock_feels2 = TweetFeels("abcd")
        mock_feels2._filter_level = 'medium'
        mock_feels2.on_data(data)
        mock_feels2._feels.insert_tweet.assert_not_called()

        # test buffer limit. no inserts until we are over limit
        mock_feels2.buffer_limit = 2
        mock_feels2.filter_level = 'low'
        mock_feels2.on_data(data)
        mock_feels2._feels.insert_tweet.assert_not_called()
        mock_feels2.on_data(data)
        mock_feels2.on_data(data)
        mock_feels._feels.insert_tweet.assert_called_once()

    def test_sentiment(self):
        mock_feels = TweetFeels("abcd")
        mock_feels._feels.tweets_since = MagicMock(return_value=[])
        mock_feels._sentiment = 0.5
        self.assertEqual(mock_feels.sentiment, 0.5)

    def test_buffer(self):
        mock_feels = TweetFeels('abcd')
        mock_feels.buffer_limit = 5
        feels_db = TweetData(file='sample.sqlite')
        mock_feels._feels = feels_db
        with open(self.tweets_data_path) as tweets_file:
            lines = list(filter(None, (line.rstrip() for line in tweets_file)))
            for line in lines[0:3]:
                t = Tweet(json.loads(line))
                mock_feels.on_data(t)
            self.assertEqual(len(mock_feels._tweet_buffer), 3)
            for line in lines[3:6]:
                t = Tweet(json.loads(line))
                mock_feels.on_data(t)
            time.sleep(1) #this waits for items to finish popping off the buffer
            self.assertEqual(len(mock_feels._tweet_buffer), 0)
            dfs = [df for df in mock_feels._feels.all]
            self.assertEqual(len(dfs[0]), 6)
        os.remove('sample.sqlite')
